/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */
import 'dart:async';
import 'dart:io';

import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/unit_test_executor.dart';
import 'package:welltested/src/executors/writer_executor.dart';
import 'package:welltested/src/extensions/annotation_ext.dart';
import 'package:welltested/src/extensions/class_element_extension.dart';
import 'package:welltested/src/extensions/extension_element_ext.dart';
import 'package:welltested/src/extensions/test_request_extension.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/repositories/writer_repository.dart';
import 'package:welltested/src/utilities/cache_manager.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';

import '../src/executors/ast_elm_executor.dart';
import '../src/utilities/typedef.dart';

class UnitGenerator {
  UnitGenerator(
      {required this.dartServerRepository, required this.cacheManager});
  final DartServerRepository dartServerRepository;
  final CacheManager cacheManager;
  late final GenerationRepository generationRepository;
  late final WriterRepository writerRepository;

  ///Executors
  late final FixerExecutor fixerExecutor;
  late final AstElementExecutor astElementExecutor;
  late final UnitTestExecutor unitTestExecutor;
  late final WriterExecutor writerExecutor;

  void init() {
    generationRepository = GenerationRepository();
    writerRepository = WriterRepository();
    fixerExecutor = FixerExecutor(dartServerRepository.server);
    astElementExecutor = AstElementExecutor();
    unitTestExecutor = UnitTestExecutor(generationRepository);
    writerExecutor =
        WriterExecutor(writerRepository, dartServerRepository, fixerExecutor);
  }

  Future<void> generateForClass({
    required ClassElement classElement,
    required ElementAnnotation annotation,
    required CompilationUnit compilationUnit,
    required ClassDeclaration classAstNode,
    required PubspecData pubspecData,
    required File classFile,
    required String? requestedFunction,
    required TestRunInfo testRunInfo,
    required bool dryrun,
    required bool testcases,
    required bool testability,
  }) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(classFile.path)}';

    final methodAstNodes = await classElement.getMethodDeclarations;

    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      final method = methodAstNodes.entries.elementAt(i);
      if (requestedFunction != null &&
          method.key.toLowerCase() != requestedFunction.trim().toLowerCase())
        continue;
      if (dryrun) {
        await _dryrunForMethod(
            method: method,
            compilationUnit: compilationUnit,
            classAstNode: classAstNode,
            fileImport: fileImport,
            excludedMethods: excludedMethods,
            className: classElement.displayName,
            methodAstNodes: methodAstNodes);
      } else if (testcases) {
        await _testcasesForMethod(
            method: method,
            compilationUnit: compilationUnit,
            classAstNode: classAstNode,
            fileImport: fileImport,
            excludedMethods: excludedMethods,
            subjectFile: classFile,
            className: classElement.displayName,
            methodAstNodes: methodAstNodes);
      } else if (testability) {
        await _testabilityForMethod(
            method: method,
            compilationUnit: compilationUnit,
            classAstNode: classAstNode,
            fileImport: fileImport,
            excludedMethods: excludedMethods,
            className: classElement.displayName,
            methodAstNodes: methodAstNodes);
      } else {
        await _generateTestsForMethod(
            method: method,
            compilationUnit: compilationUnit,
            classAstNode: classAstNode,
            fileImport: fileImport,
            excludedMethods: excludedMethods,
            subjectFile: classFile,
            testRunInfo: testRunInfo,
            className: classElement.displayName,
            methodAstNodes: methodAstNodes);
      }
    }
  }

  Future<void> generateForExtension({
    required ExtensionElement extElement,
    required ElementAnnotation annotation,
    required CompilationUnit compilationUnit,
    required ExtensionDeclaration extAstNode,
    required PubspecData pubspecData,
    required File extFile,
    required String? requestedFunction,
    required TestRunInfo testRunInfo,
    required bool dryrun,
    required bool testcases,
    required bool testability,
  }) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(extFile.path)}';

    final methodAstNodes = await extElement.getMethodDeclarations;

    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      final method = methodAstNodes.entries.elementAt(i);
      if (requestedFunction != null &&
          method.key.toLowerCase() != requestedFunction.trim().toLowerCase())
        continue;
      if (dryrun) {
        await _dryrunForMethod(
            method: method,
            compilationUnit: compilationUnit,
            extAstNode: extAstNode,
            fileImport: fileImport,
            excludedMethods: excludedMethods,
            className: extElement.displayName,
            methodAstNodes: methodAstNodes);
      } else if (testability) {
        await _testabilityForMethod(
            method: method,
            compilationUnit: compilationUnit,
            extAstNode: extAstNode,
            fileImport: fileImport,
            excludedMethods: excludedMethods,
            className: extElement.displayName,
            methodAstNodes: methodAstNodes);
      } else if (testcases) {
        await _testcasesForMethod(
            method: method,
            compilationUnit: compilationUnit,
            extAstNode: extAstNode,
            fileImport: fileImport,
            excludedMethods: excludedMethods,
            subjectFile: extFile,
            className: extElement.displayName,
            methodAstNodes: methodAstNodes);
      } else {
        await _generateTestsForMethod(
            method: method,
            compilationUnit: compilationUnit,
            extAstNode: extAstNode,
            fileImport: fileImport,
            excludedMethods: excludedMethods,
            subjectFile: extFile,
            testRunInfo: testRunInfo,
            className: extElement.displayName,
            methodAstNodes: methodAstNodes);
      }
    }
  }

  Future<void> generateForGlobalExeculatbleElement({
    required ExecutableElement methodElement,
    required FunctionDeclaration methodAstNode,
    required CompilationUnit compilationUnit,
    required ElementAnnotation annotation,
    required PubspecData pubspecData,
    required File methodFile,
    required TestRunInfo testRunInfo,
    required bool dryrun,
    required bool testcases,
    required bool testability,
  }) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(methodFile.path)}';
    if (dryrun) {
      await _dryrunForMethod(
        method: MapEntry(methodElement.displayName, methodAstNode),
        compilationUnit: compilationUnit,
        fileImport: fileImport,
        excludedMethods: excludedMethods,
      );
    } else if (testability) {
      await _testabilityForMethod(
        method: MapEntry(methodElement.displayName, methodAstNode),
        compilationUnit: compilationUnit,
        fileImport: fileImport,
        excludedMethods: excludedMethods,
      );
    } else if (testcases) {
      await _testcasesForMethod(
        method: MapEntry(methodElement.displayName, methodAstNode),
        compilationUnit: compilationUnit,
        fileImport: fileImport,
        subjectFile: methodFile,
        excludedMethods: excludedMethods,
      );
    } else {
      await _generateTestsForMethod(
          method: MapEntry(methodElement.displayName, methodAstNode),
          compilationUnit: compilationUnit,
          fileImport: fileImport,
          excludedMethods: excludedMethods,
          subjectFile: methodFile,
          testRunInfo: testRunInfo);
    }
  }

  Future<void> generateForGlobalMethod({
    required FunctionElement methodElement,
    required FunctionDeclaration methodAstNode,
    required CompilationUnit compilationUnit,
    required ElementAnnotation annotation,
    required PubspecData pubspecData,
    required File methodFile,
    required TestRunInfo testRunInfo,
    required bool dryrun,
    required bool testcases,
    required bool testability,
  }) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(methodFile.path)}';

    if (dryrun) {
      await _dryrunForMethod(
        method: MapEntry(methodElement.displayName, methodAstNode),
        compilationUnit: compilationUnit,
        fileImport: fileImport,
        excludedMethods: excludedMethods,
      );
    } else if (testability) {
      await _testabilityForMethod(
        method: MapEntry(methodElement.displayName, methodAstNode),
        compilationUnit: compilationUnit,
        fileImport: fileImport,
        excludedMethods: excludedMethods,
      );
    } else if (testcases) {
      await _testcasesForMethod(
        method: MapEntry(methodElement.displayName, methodAstNode),
        compilationUnit: compilationUnit,
        fileImport: fileImport,
        excludedMethods: excludedMethods,
        subjectFile: methodFile,
      );
    } else {
      await _generateTestsForMethod(
          method: MapEntry(methodElement.displayName, methodAstNode),
          compilationUnit: compilationUnit,
          fileImport: fileImport,
          excludedMethods: excludedMethods,
          subjectFile: methodFile,
          testRunInfo: testRunInfo);
    }
  }

  Future<void> _generateTestsForMethod({
    required MapEntry<String, Declaration> method,
    required CompilationUnit compilationUnit,
    required String fileImport,
    required List<String> excludedMethods,
    required File subjectFile,
    required TestRunInfo testRunInfo,
    ClassDeclaration? classAstNode,
    ExtensionDeclaration? extAstNode,
    String? className,
    Map<String, MethodDeclaration>? methodAstNodes,
  }) async {
    try {
      final testRequestData = await unitTestExecutor.prepareRequest(
        method: method,
        compilationUnit: compilationUnit,
        classAstNode: classAstNode,
        extensionAstNode: extAstNode,
        fileImport: fileImport,
        className: className,
        methodAstNodes: methodAstNodes,
        excludedMethods: excludedMethods,
      );
      if (testRequestData == null) return;
      wtLog.updateSpinnerMessage(
          "Generating Test for ${testRequestData.functionName}");
      final file =
          PathUtils.correspondingPermanentTestFile(method.key, subjectFile);

      final existStatus = await cacheManager.testExistStatus(
          testRequestData.codeInScrutity(excludeAnnotations: false),
          subjectFile.path,
          testRequestData.functionName,
          className: testRequestData.parentEntityName);
      bool update = false;
      late TestGenerationStatus testGenerationStatus;
      switch (existStatus) {
        case TestExistStatus.generationExists:
          testGenerationStatus = TestGenerationStatus.omit;
          return;
        case TestExistStatus.onlyRequestExists:
          testRequestData.forceGenerate = true;
          testGenerationStatus = TestGenerationStatus.forceUpdate;
          break;
        case TestExistStatus.outdatedCodeExists:
          update = true;
          testGenerationStatus = TestGenerationStatus.update;
          break;
        case TestExistStatus.doesNotExist:
          testGenerationStatus = TestGenerationStatus.create;
          break;
      }

      final unitTests = await unitTestExecutor.buildUnitTests(testRequestData);
      String? completion = unitTests['fixed_completion'];
      late File testFile;
      if (completion != null) {
        testFile = await writerExecutor.outputGeneration(file.path, completion);
      } else {
        completion = unitTests['model_completion'];
        if (completion == null) {
          wtLog.warning(
              "⛔️ Sorry, we can't process ${testRequestData.functionName} at the moment. Please try again Later.");
          return;
        }
        testFile = await writerExecutor.processGeneration(
            file.path, completion, method.key);
      }
      await cacheManager.cacheTest(
          code: testRequestData.codeInScrutity(excludeAnnotations: false),
          codeFilePath: subjectFile.path,
          test: completion,
          testFilePath: testFile.path,
          className: testRequestData.parentEntityName,
          functionName: testRequestData.functionName);
      testRunInfo.incrementFromAction(testGenerationStatus);
      wtLog.log(
          '✔️ Tests ${update ? 'updated' : 'generated'} for ${method.key}');
    } catch (e, stackTrace) {
      testRunInfo.incrementFromAction(TestGenerationStatus.failed);
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.warning("Error generating unit tests for method: ${method.key}");
      wtLog.log(e.toString());
    }
  }

  Future<void> _dryrunForMethod({
    required MapEntry<String, Declaration> method,
    required CompilationUnit compilationUnit,
    required String fileImport,
    required List<String> excludedMethods,
    ClassDeclaration? classAstNode,
    ExtensionDeclaration? extAstNode,
    String? className,
    Map<String, MethodDeclaration>? methodAstNodes,
  }) async {
    try {
      final testRequestData = await unitTestExecutor.prepareRequest(
        method: method,
        compilationUnit: compilationUnit,
        classAstNode: classAstNode,
        extensionAstNode: extAstNode,
        fileImport: fileImport,
        className: className,
        methodAstNodes: methodAstNodes,
        excludedMethods: excludedMethods,
      );
      if (testRequestData == null) return;
      final stream = await unitTestExecutor.dryRun(testRequestData);
      stream.listen((event) {
        if (event.isEmpty) return;
        event = event.replaceAll("\n", '');
        if (testRequestData.testcases.isEmpty) {
          _handleSuggestions(event);
        } else {
          _handleTestcases(event);
        }
      }).onDone(() {
        wtLog.stopSpinner(
            message: "✅ Succeeded. Dry Run completed for ${method.key}");
      });
    } catch (e, stackTrace) {
      wtLog.stopSpinner(
          message: "Oops. Dry Run failed for ${method.key}",
          severity: MessageSeverity.Error);
      wtLog.log(e.toString());
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
    }
  }

  void _handleSuggestions(String input) {
    final severity = _parseSeverity(input);
    if (severity == Criticality.None) {
      wtLog.log(input);
    } else if (severity == Criticality.critical) {
      wtLog.error(input);
    } else if (severity == Criticality.warning) {
      wtLog.warning(input);
    } else if (severity == Criticality.info) {
      wtLog.info(input);
    }
  }

  void _handleTestcases(String input) {
    final inputs = input.split("#");
    for (String input in inputs.where((element) => element.isNotEmpty)) {
      input = "#$input".replaceAll("\\n", "\n");
      if (input.startsWith('# Testcase:') || input.startsWith("# Check:")) {
        wtLog.info(input);
        wtLog.startSpinner('Evaluating');
      } else if (input.startsWith('# Result:')) {
        final result = input.replaceFirst('# Result:', '').trim().toLowerCase();
        if (result == 'pass') {
          wtLog.success(input);
        } else if (result == 'fail') {
          wtLog.error(input);
        } else {
          wtLog.log(input);
        }
      } else if (input.startsWith('# Actionables:')) {
        // check if the actionables is None
        final actionables = input
            .replaceFirst('# Actionables:', '')
            .trim()
            .toLowerCase()
            .replaceAll(' ', '');
        if (actionables.toLowerCase() != 'none') {
          wtLog.log(input + "\n");
        } else {
          print("");
        }
      } else if (input.startsWith("# Evaluation:")) {
        // Stop the spinner started for evaluating
        wtLog.stopSpinner();
      } else {
        wtLog.log(input);
      }
    }
  }

  Criticality _parseSeverity(String input) {
    final inputs = input.split("#");
    for (String input in inputs.where((element) => element.isNotEmpty)) {
      input = "#$input";
      if (input.contains('[Critical]')) {
        return Criticality.critical;
      } else if (input.contains('[Warning]')) {
        return Criticality.warning;
      } else if (input.contains('[Info]')) {
        return Criticality.info;
      } else {
        return Criticality.None;
      }
    }
    return Criticality.None;
  }

  Future<void> _testcasesForMethod({
    required MapEntry<String, Declaration> method,
    required CompilationUnit compilationUnit,
    required String fileImport,
    required File subjectFile,
    required List<String> excludedMethods,
    ClassDeclaration? classAstNode,
    ExtensionDeclaration? extAstNode,
    String? className,
    Map<String, MethodDeclaration>? methodAstNodes,
  }) async {
    try {
      final testRequestData = await unitTestExecutor.prepareRequest(
        method: method,
        compilationUnit: compilationUnit,
        classAstNode: classAstNode,
        extensionAstNode: extAstNode,
        fileImport: fileImport,
        className: className,
        methodAstNodes: methodAstNodes,
        excludedMethods: excludedMethods,
      );
      if (testRequestData == null) return;
      final response = await unitTestExecutor.testcases(testRequestData);
      final testcases = List<String>.from(response['testcases']);
      String currentCode = await subjectFile.readAsString();
      final existingTestcases = method.value.metadata.testcasesAnnotations;
      currentCode = currentCode.substring(
              0, method.value.firstTokenAfterCommentAndMetadata.offset - 1) +
          '@Testcases(${testcases.map((e) {
            late String value;
            if (e.contains("'")) {
              value =
                  "\"$e\""; //double quotes when string contains single quotes
            } else {
              value = "'$e'";
            }
            if (testcases.lastIndexOf(e) == testcases.length - 1) {
              value = value +
                  ','; //append comma at the end to avoid dart analyzer warning
            }
            return value;
          }).toList()})' +
          '\n' +
          currentCode
              .substring(method.value.firstTokenAfterCommentAndMetadata.offset);

      if (existingTestcases.isNotEmpty) {
        final testcaseOffsets = existingTestcases
            .map((e) => [e.offset, e.length])
            .toList()
            .reversed
            .toList();
        for (final offset in testcaseOffsets) {
          currentCode =
              currentCode.replaceRange(offset[0], offset[0] + offset[1], '');
        }
      }
      await writerExecutor.outputGeneration(subjectFile.path, currentCode);
      wtLog.stopSpinner(
          message: "✅ Succeeded. Testcases generated for ${method.key}");
    } catch (e, stackTrace) {
      wtLog.stopSpinner(
          message: "Oops. Testcases generation failed for ${method.key}",
          severity: MessageSeverity.Error);
      wtLog.log(e.toString());
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
    }
  }

  Future<void> _testabilityForMethod({
    required MapEntry<String, Declaration> method,
    required CompilationUnit compilationUnit,
    required String fileImport,
    required List<String> excludedMethods,
    ClassDeclaration? classAstNode,
    ExtensionDeclaration? extAstNode,
    String? className,
    Map<String, MethodDeclaration>? methodAstNodes,
  }) async {
    try {
      final testRequestData = await unitTestExecutor.prepareRequest(
        method: method,
        compilationUnit: compilationUnit,
        classAstNode: classAstNode,
        extensionAstNode: extAstNode,
        fileImport: fileImport,
        className: className,
        methodAstNodes: methodAstNodes,
        excludedMethods: excludedMethods,
      );
      if (testRequestData == null) return;
      final stream = await unitTestExecutor.testability(testRequestData);
      stream.listen((event) {
        _handleTestcases(event);
      }).onDone(() {
        wtLog.stopSpinner(
            message:
                "✅ Succeeded. Testability check completed for ${method.key}");
      });
    } catch (e, stackTrace) {
      wtLog.stopSpinner(
          message: "Oops. Dry Run failed for ${method.key}",
          severity: MessageSeverity.Error);
      wtLog.log(e.toString());
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
    }
  }
}
