/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */
import 'dart:io';

import 'package:analysis_server_lib/analysis_server_lib.dart';
import 'package:analysis_server_lib/analysis_server_lib.dart' as serverlib;
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/src/extensions/analysis_errors_extenstion.dart';
import 'package:welltested/src/extensions/list_extension.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';

class FixerExecutor {
  final serverlib.AnalysisServer server;
  FixerExecutor(this.server);

  Future<void> setRoots(List<String> includedFiles) async {
    await server.analysis.setAnalysisRoots(includedFiles, []);
  }

  Future<bool> hasCompilationErrors(String filePath) async {
    await server.analysis.setAnalysisRoots([filePath],
        []); //TODO: Shift to Dart Server repository add files to root.
    final analysis = await server.analysis.getErrors(filePath);
    bool hasErrors = false;
    for (var error in analysis.errors) {
      if (error.severity == 'ERROR') {
        hasErrors = true;
        break;
      }
    }
    return hasErrors;
  }

  Future<String> findAndFixIssues(
      String filePath, String functionIdentifer) async {
    File file = File(filePath);
    final String fileContent = await file.readAsString();
    late String outputContent;
    try {
      // getErrors is failing for one file in every run.
      // root cause needs to be fixed.
      final analysis = await server.analysis.getErrors(filePath);
      outputContent = await _fixImportIssues(
          analysis.errors.fixableUndefinedErrors, filePath, fileContent, file);
    } catch (e, stackTrace) {
      outputContent = fileContent;
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
    }
    outputContent = addMocksFile(outputContent, functionIdentifer);

    return outputContent;
  }

  String addMocksFile(String content, String functionidentifier) {
    if (!content.contains("@GenerateMocks([")) {
      return content;
    }
    return content.replaceRange(
        0, 0, "import '$functionidentifier.welltested_test.mocks.dart';");
  }

  Future<String> _fixImportIssues(List<AnalysisError> undefinedErrors,
      String filePath, String fileContent, File file) async {
    final List<SourceEdit> replacements = [];
    for (final error in undefinedErrors) {
      final FixesResult fix = await server.edit.getFixes(
        filePath,
        error.location.offset,
      );

      if (fix.fixes.isNotEmpty) {
        Set<SourceEdit> _fixes = Set();
        for (var element in fix.fixes) {
          for (var e in element.fixes) {
            if (e.id?.contains("dart.fix.import") ?? false) {
              _fixes.add(e.edits.first.edits.first);
            }
          }
        }
        replacements.addAll(_fixes);
      }
    }
    final updatedContent = _runSourceEditSequence(fileContent, replacements);
    await file.writeAsString(updatedContent, flush: true);
    return updatedContent;
  }

  String _runSourceEditSequence(String content, List<SourceEdit> sourceEdits) {
    sourceEdits.sort((a, b) => a.offset.compareTo(b.offset));
    sourceEdits = sourceEdits.filterDuplicationReplacements;
    String updatedContent = content;
    int shift = 0;
    for (final edit in sourceEdits) {
      updatedContent = updatedContent.replaceRange(edit.offset + shift,
          edit.offset + edit.length + shift, edit.replacement);
      shift = shift + edit.replacement.length - edit.length;
    }
    return updatedContent;
  }
}
