/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */

import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/src/extensions/string_extensions.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';
import 'package:welltested/src/utilities/utilities.dart';

import '../utilities/typedef.dart';

extension ClassElementExt on ClassElement {
  bool get isSimple {
    return !this.isPrivate &&
        !this.hasProtected &&

        ///TODO: Replace [!this.isAbstract] with [this.isConstructable] after analyzer 5.7.1
        !this.isAbstract &&
        this.constructors.indexWhere((e) => !e.isFactory) !=
            -1; //Has non factory constructors
  }

  ///Contains elements that aren't part of core data type
  List<ClassElement> get nonDartElements {
    List<ClassElement> nonSDK = [];
    this.fields.forEach((field) {
      final fieldElement = field.type.element;
      if (fieldElement is ClassElement) {
        if (!fieldElement.library.isInSdk) {
          nonSDK.add(fieldElement);
        }
      }
    });
    return nonSDK;
  }

  ///Returns the source code of the class without factory constructors and methods.
  Future<String> simplefiedSource() async {
    String classCode = (await getResolvedAstNodeForElement(this)).toSource();
    if (classCode.length < 1800) return classCode;
    if (classCode.length < 2500) return classCode.shrink;
    //remove contructor and nested methods.
    for (final constructor in this.constructors) {
      if (constructor.isFactory) {
        final replaceString =
            (await getResolvedAstNodeForElement(constructor)).toSource();
        classCode = classCode.replaceAll(replaceString, '');
      }
    }

    for (final method in this.methods) {
      final replaceString =
          (await getResolvedAstNodeForElement(method)).toSource();
      classCode = classCode.replaceAll(replaceString, '');
    }
    if (classCode.length < 2200) return classCode;
    return classCode.shrink;
  }

  bool get checkIfAWidgetOrState {
    try {
      final superTypes = this
          .allSupertypes
          .map((supertype) => supertype.element.name)
          .toList();
      //[ConsumerStatefulWidget, StatefulWidget, Widget, DiagnosticableTree, Object, Diagnosticable]
      final widgetIndex =
          superTypes.indexWhere((element) => element == 'Widget');
      if (widgetIndex != -1) {
        if (superTypes[widgetIndex + 1] == 'DiagnosticableTree') return true;
      }
      //[ConsumerState<ConsumerStatefulWidget>, State<ConsumerStatefulWidget>, Object, Diagnosticable]
      final stateIndex = superTypes.indexWhere((element) => element == 'State');
      if (stateIndex != -1) {
        if (superTypes[stateIndex + 1] == 'Object') return true;
        if (superTypes[stateIndex + 2] == 'Diagnosticable') return true;
      }
      return false;
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      return false;
    }
  }

  Future<Map<String, MethodDeclaration>> get getMethodDeclarations async {
    final Map<String, MethodDeclaration> methodAstNodes = {};
    for (final method in methods) {
      final node =
          (await getResolvedAstNodeForElement(method)) as MethodDeclaration;
      methodAstNodes[method.name] = node;
    }
    return methodAstNodes;
  }
}

extension ClassDeclarationExtension on ClassDeclaration {
  List<AstNode>? checkIfWidget() {
    if (this.extendsClause?.superclass.name == 'StatelessWidget' ||
        this.extendsClause?.superclass.name.name == 'ConsumerStatefulWidget' ||
        this.extendsClause?.superclass.name.name == 'StatefulWidget' ||
        (this.extendsClause?.superclass.name.name.contains('GetWidget') ??
            false)) {
      return parent?.childEntities
          .where((element) {
            return element is ImportDirective ||
                element == this ||
                (element is ClassDeclaration &&
                    (element.extendsClause?.superclass.name.name
                            .contains('State') ??

                        ///TODO: Consider widgets with other extending classes
                        false));
          })
          .map((e) => e as AstNode)
          .toList();
    } else
      return null;
  }
}

extension Processing on List<ClassElement> {
  Future<List> processModels() async {
    final List<ClassElement> complexModels = [];
    final List<ClassElement> simpleModels = [];

    for (final element in this) {
      if (!element.isSimple) {
        complexModels.add(element);
        continue;
      }
      final nonSDKElements = element.nonDartElements;
      if (nonSDKElements.isEmpty) {
        simpleModels.add(element);
        continue;
      }

      bool isComplexModel = false;
      for (final nonSDKElement in nonSDKElements) {
        final moreNonSDKElements = nonSDKElement.nonDartElements;
        if (moreNonSDKElements.isNotEmpty) {
          isComplexModel = true;
          break;
        }
      }
      if (isComplexModel) {
        complexModels.add(element);
      } else {
        //TODO: Maybe define a max field limit length of comined models
        simpleModels.add(element);
        simpleModels.addAll(nonSDKElements);
      }
    }
    final List<String> simpleModelCode = [];

    for (final simpleModel in simpleModels) {
      simpleModelCode.add(await simpleModel.simplefiedSource());
    }

    return [simpleModelCode, complexModels];
  }
}

extension IntegrationsTestExtensions on ClassElement {
  Future<String> setup(ElementToNode elementToNode) async {
    final descriptionElm = fields.where((field) => field.name == 'setup').first;
    final descriptionNode = ((await elementToNode(
      descriptionElm.getter!.nonSynthetic,
    ) as MethodDeclaration)
            .body as ExpressionFunctionBody)
        .expression;
    final description = (descriptionNode as SimpleStringLiteral).value;
    return description;
  }

  Future<List<String>> testcases(ElementToNode elementToNode) async {
    final testCasesElm =
        fields.where((field) => field.name == 'testcases').first;
    final testCasesNode =
        ((await elementToNode(testCasesElm.getter!.nonSynthetic)
                    as MethodDeclaration)
                .body as ExpressionFunctionBody)
            .expression;
    final testCases = (testCasesNode as ListLiteral)
        .elements
        .map((e) => (e as SimpleStringLiteral).value)
        .toList();
    return testCases;
  }
}
