/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */
import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:http/http.dart' as http;
import 'package:http/http.dart';
import 'package:http/io_client.dart';
import 'package:welltested/src/models/test_request/base_test_request.dart';
import 'package:welltested/src/utilities/env_parser.dart';

class GenerationRepository {
  final baseUrl = 'https://api.welltested.ai';

  Future<Map<String, dynamic>> buildUnitTests(
      UnitTestRequest testRequestData) async {
    try {
      final DotEnv env = DotEnv(includePlatformEnvironment: true)..load();
      final apiKey = env.getOrElse("WELLTESTED_API", () => "");
      if (apiKey.isEmpty) {
        throw "🚨 WELLTESTED_API key not found. Please refer documentation for more Info.";
      }
      final response = await http.post(
          Uri.parse('$baseUrl/ai/unit-tests/generate'),
          body: testRequestData.toJson(),
          headers: {"apikey": apiKey});
      if (response.statusCode == 200) {
        return jsonDecode(response.body);
      } else {
        throw (_handleErrorStatusCodes(response));
      }
    } catch (e) {
      throw (e);
    }
  }

  Future<Map<String, dynamic>> updateTest(
      UnitTestRequest testRequestData, String fixedTest) async {
    try {
      final DotEnv env = DotEnv(includePlatformEnvironment: true)..load();
      final apiKey = env.getOrElse("WELLTESTED_API", () => "");
      if (apiKey.isEmpty) {
        throw "🚨 WELLTESTED_API key not found. Please refer documentation for more Info.";
      }

      final response = await http.post(
          Uri.parse('$baseUrl/ai/unit-tests/update'),
          body: testRequestData.toJson()..['fixed_test_case'] = fixedTest,
          headers: {"apikey": apiKey});
      if (response.statusCode == 200) {
        return jsonDecode(response.body);
      } else {
        throw (_handleErrorStatusCodes(response));
      }
    } catch (e) {
      throw (e);
    }
  }

  Future<Map<String, dynamic>> testcases(
      UnitTestRequest testRequestData) async {
    try {
      final DotEnv env = DotEnv(includePlatformEnvironment: true)..load();
      final apiKey = env.getOrElse("WELLTESTED_API", () => "");
      if (apiKey.isEmpty) {
        throw "🚨 WELLTESTED_API key not found. Please refer documentation for more Info.";
      }
      final response = await http.post(
          Uri.parse('$baseUrl/ai/unit-tests/test-cases'),
          body: testRequestData.toJson(),
          headers: {"apikey": apiKey});
      if (response.statusCode == 200) {
        return jsonDecode(response.body);
      } else {
        throw (_handleErrorStatusCodes(response));
      }
    } catch (e) {
      throw (e);
    }
  }

  Stream<String> dryRun(UnitTestRequest testRequestData) async* {
    try {
      final DotEnv env = DotEnv(includePlatformEnvironment: true)..load();
      final apiKey = env.getOrElse("WELLTESTED_API", () => "");
      if (apiKey.isEmpty) {
        throw "🚨 WELLTESTED_API key not found. Please refer documentation for more Info.";
      }
      final url = '$baseUrl/ai/unit-tests/dry-run';

      HttpClient httpClient = HttpClient();
      IOClient ioClient = IOClient(httpClient);
      Uri uri = Uri.parse(url);
      final request = http.Request("POST", uri);
      request.bodyFields = testRequestData.toJson().cast<String, String>();
      request.headers['apikey'] = apiKey;
      final response = await ioClient.send(request);
      if (response.statusCode == 200) {
        final transformer = StreamTransformer<Uint8List, String>.fromHandlers(
          handleData: (data, sink) {
            String text = utf8.decoder.convert(data);
            List<String> lines = text.split('\n');
            for (String line in lines) {
              sink.add(line);
            }
          },
        );
        await for (String line
            in response.stream.cast<Uint8List>().transform(transformer)) {
          yield line;
        }
      } else {
        if (response.statusCode == 403) {
          throw ("🚨 Invalid Welltested API Key. Use the key recieved in the registered mail.");
        } else {
          //TODO: Replace this with a clear response from backend.
          throw ("Unknown Exception. Please reach out to team@welltested.ai for resolution");
        }
      }
    } catch (e) {
      throw (e);
    }
  }

  Stream<String> checkTestability(UnitTestRequest testRequestData) async* {
    try {
      final DotEnv env = DotEnv(includePlatformEnvironment: true)..load();
      final apiKey = env.getOrElse("WELLTESTED_API", () => "");
      if (apiKey.isEmpty) {
        throw "🚨 WELLTESTED_API key not found. Please refer documentation for more Info.";
      }
      final url = '$baseUrl/ai/unit-tests/testability';

      HttpClient httpClient = HttpClient();
      IOClient ioClient = IOClient(httpClient);
      Uri uri = Uri.parse(url);
      final request = http.Request("POST", uri);
      request.bodyFields = testRequestData.toJson().cast<String, String>();
      request.headers['apikey'] = apiKey;
      final response = await ioClient.send(request);
      if (response.statusCode == 200) {
        final transformer = StreamTransformer<Uint8List, String>.fromHandlers(
          handleData: (data, sink) {
            String text = utf8.decoder.convert(data);
            List<String> lines = text.split('\n');
            for (String line in lines) {
              sink.add(line);
            }
          },
        );
        await for (String line
            in response.stream.cast<Uint8List>().transform(transformer)) {
          yield line;
        }
      } else {
        if (response.statusCode == 403) {
          throw ("🚨 Invalid Welltested API Key. Use the key recieved in the registered mail.");
        } else {
          //TODO: Replace this with a clear response from backend.
          throw ("Unknown Exception. Please reach out to team@welltested.ai for resolution");
        }
      }
    } catch (e) {
      throw (e);
    }
  }

  //TODO: Check Error Scenarios
  String _handleErrorStatusCodes(Response response) {
    if (response.statusCode == 403) {
      return "🚨 Invalid Welltested API Key. Use the key recieved in the registered mail.";
    } else if (response.statusCode == 500) {
      return "🚨 Error: ${jsonDecode(response.body)['error']}";
    } else {
      return "Unknown Exception. Please reach out to team@welltested.ai for resolution. Details: ${response.body}";
    }
  }
}
