import 'dart:io';

import 'package:args/command_runner.dart';
import 'package:path/path.dart' as p;
import 'package:welltested/cli/operation.dart';
import 'package:welltested/src/extensions/pubspec_data_extension.dart';
import 'package:welltested/src/models/dependency_info.dart';
import 'package:welltested/src/utilities/env_parser.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/pubspec_parser.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';
import 'package:welltested/src/utilities/utilities.dart';

import 'cache_manager.dart';

class UnitCommand extends Command {
  bool dryrun = false;
  bool testcases = false;
  bool testability = false;
  UnitCommand(String parent) {
    argParser.addFlag('verbose', abbr: 'v', callback: (value) {
      wtLog.verbose = value;
    });
    argParser.addFlag('dry-run', callback: (value) {
      this.dryrun = value;
    });
    argParser.addFlag('cases', callback: (value) {
      this.testcases = value;
    });
    argParser.addFlag('testability', callback: (value) {
      this.testability = value;
    });
    if (parent == 'generate') {
      argParser.addOption('file',
          abbr: 'f',
          help: 'relative path of the file',
          valueHelp: 'bloc/login_bloc.dart');
      argParser.addOption("class",
          abbr: "c",
          help: 'name of the annotated class',
          valueHelp: 'LoginBloc');
      argParser.addOption("method",
          abbr: 'm',
          help: "name of the method under annotated class",
          valueHelp: 'loginWithEmailAndPassword');
    }
  }
  @override
  String get name => 'unit';

  @override
  String get description => 'Subcommand to select unit test';
  Future<void> run() async {
    if (verify_two_true([dryrun, testcases])) {
      wtLog.error(
          'No two of flags `dryrun` and `testcases` can be passed together');
      return;
    }
    if (this.dryrun || this.testcases || this.testability) {
      final methodName = argResults?['method'];
      if (methodName == null) {
        wtLog.error("Method name is required for dry run.");
        return;
      }
    }

    String currentDirectory = PathUtils.currentPath;
    final pubspecData = await PubspecParser.forPath(currentDirectory);
    setPackage(pubspecData.welltestedInfo?.dependencyVersion,
        pubspecData.welltestedInfo?.dev);
    await CacheManager().initialize();

    if (parent?.name == 'generate') {
      String? filePath = argResults?['file'];

      // if file path is given, add the current directory to it
      if (filePath != null) {
        filePath = p.join(PathUtils.currentPath, filePath);
        if (!File(filePath).existsSync()) {
          wtLog.error("File does not exist: $filePath");
          return;
        }
      }
      await CLIOperation().generateUnit(
          pubspecData: pubspecData,
          filePath: filePath,
          className: argResults?['class'],
          methodName: argResults?['method'],
          dryrun: dryrun,
          testcases: testcases,
          testability: testability);
      ;
    } else if (parent?.name == 'save') {
      await CLIOperation.saveUnit(pubspecData: pubspecData);
    }
    await CacheManager().closeAndSettle();
  }
}

class GenerateCommmand extends Command {
  @override
  String get name => 'generate';
  @override
  String get description => 'Generate tests for the annotated code';

  GenerateCommmand() {
    addSubcommand(UnitCommand(name));
  }
}

class SaveCommand extends Command {
  @override
  String get name => 'save';

  @override
  String get description => 'Save updates made to existing tests';

  SaveCommand() {
    addSubcommand(UnitCommand(name));
  }
}

class InitCommand extends Command {
  @override
  String get name => 'init';

  @override
  String get description => 'Initialize welltested in your project';

  @override
  Future<void> run() async {
    try {
      final pubspecData = await PubspecParser.forPath(PathUtils.currentPath);
      await _addDependency(
          "welltested_annotation", pubspecData.welltestedAnnotationInfo,
          isDev: false, isFlutter: pubspecData.isFlutter);
      await _addDependency("build_runner", pubspecData.build_runnerInfo,
          isDev: true, isFlutter: pubspecData.isFlutter);
      await _addDependency("mockito", pubspecData.mockitoInfo,
          isDev: true, isFlutter: pubspecData.isFlutter);
      if (pubspecData.isFlutter) {
        await _addDependency("flutter_test", pubspecData.flutterTestInfo,
            isDev: true, isSDK: true, isFlutter: true);
      } else {
        await _addDependency("test", pubspecData.testInfo,
            isDev: true, isSDK: false, isFlutter: false);
      }
      await _addKeyToEnv();
      wtLog.info("Welltested initialized ✅");
    } catch (e) {
      wtLog.error("Error initializing welltested");
    }
  }

  _addDependency(String dependency, DependencyInfo? dependencyInfo,
      {bool isDev = false, bool isSDK = false, bool isFlutter = false}) async {
    final command = isFlutter ? "flutter" : "dart";

    if (dependencyInfo == null) {
      await runCommand(
          command,
          ["pub", "add", dependency]
            ..addAll(isSDK ? ["--sdk=flutter"] : [])
            ..addAll(isDev ? ["--dev"] : []));
    } else if (isDev && !dependencyInfo.dev) {
      await runCommand(command, ["pub", "remove", dependency]);
      await runCommand(
          command,
          ["pub", "add", dependency]
            ..addAll(isSDK ? ["--sdk=flutter"] : [])
            ..add(isDev ? "--dev" : ""));
      ;
    }
  }

  _addKeyToEnv() async {
    final DotEnv env = DotEnv(includePlatformEnvironment: false)..load();
    final apiKey = env.getOrElse("WELLTESTED_API", () => "");
    if (apiKey.isEmpty) {
      wtLog.info("Please enter your welltested api key received in email");
      stdout.write(">> ");
      String? key = stdin.readLineSync();
      if (key == null || key.isEmpty) {
        wtLog.error("Api key can't be empty");
        return;
      }
      env.addNew({"WELLTESTED_API": key});
      wtLog.info("API Key successfully added.");
    }
  }
}
