/// Permission is hereby granted, free of charge, to any person obtaining a copy
/// of this env_parser.dart which is extracted from dotenv package by mockturtl
/// and used directly to support versioning requirement of the package.

import 'dart:io';

class DotEnv {
  final bool includePlatformEnvironment;

  final _map = <String, String>{};

  DotEnv({this.includePlatformEnvironment = false}) {
    if (includePlatformEnvironment) _addPlatformEnvironment();
  }

  Map<String, String> get map => _map;

  String? operator [](String key) => _map[key];

  void addAll(Map<String, String> other) => _map.addAll(other);

  void addNew(Map<String, String> other) {
    String _stringToAdd = '';
    for (var key in other.keys) {
      if (!_map.containsKey(key)){
        _stringToAdd += '${_map.isEmpty?'':'\n'}$key=${other[key]}';
      }
    }
    File('.env').writeAsStringSync(_stringToAdd, mode: FileMode.append);
  }

  void clear() {
    _map.clear();
    if (includePlatformEnvironment) _addPlatformEnvironment();
  }

  String getOrElse(String key, String Function() orElse) =>
      isDefined(key) ? _map[key]! : orElse();

  bool isDefined(String key) => _map[key]?.isNotEmpty ?? false;

  bool isEveryDefined(Iterable<String> vars) => vars.every(isDefined);

  void load(
      [Iterable<String> filenames = const ['.env'],
      Parser psr = const Parser()]) {
    for (var filename in filenames) {
      var f = File.fromUri(Uri.file(filename));
      var lines = _verify(f);
      _map.addAll(psr.parse(lines));
    }
  }

  void _addPlatformEnvironment() => _map.addAll(Platform.environment);

  List<String> _verify(File f) {
    if (!f.existsSync()) {
      stderr.writeln('[dotenv] Load failed: file not found: $f');
      return [];
    }
    return f.readAsLinesSync();
  }
}

class Parser {
  static const _singleQuot = "'";
  static const _keyword = 'export';

  static final _comment = new RegExp(r'''#.*(?:[^'"])$''');
  static final _surroundQuotes = new RegExp(r'''^(['"])(.*)\1$''');
  static final _bashVar =
      new RegExp(r'(?:\\)?(\$)(?:{)?([a-zA-Z_][\w]*)+(?:})?');

  const Parser();

  String interpolate(String val, Map<String, String> env) =>
      val.replaceAllMapped(_bashVar, (m) {
        var k = m.group(2)!;
        return (!_has(env, k)) ? _tryPlatformEnv(k) ?? '' : env[k] ?? '';
      });

  Map<String, String> parse(Iterable<String> lines) {
    var out = <String, String>{};
    lines.forEach((line) {
      var kv = parseOne(line, env: out);
      if (kv.isEmpty) return;
      out[kv.keys.single] = kv.values.single;
    });
    return out;
  }

  Map<String, String> parseOne(String line,
      {Map<String, String> env = const {}}) {
    var stripped = strip(line);
    if (!_isValid(stripped)) return {};

    var idx = stripped.indexOf('=');
    var lhs = stripped.substring(0, idx);
    var k = swallow(lhs);
    if (k.isEmpty) return {};

    var rhs = stripped.substring(idx + 1, stripped.length).trim();
    var quotChar = surroundingQuote(rhs);
    var v = unquote(rhs);

    if (quotChar == _singleQuot) // skip substitution in single-quoted values
      return {k: v};

    return {k: interpolate(v, env)};
  }

  String strip(String line) => line.replaceAll(_comment, '').trim();

  String surroundingQuote(String val) {
    if (!_surroundQuotes.hasMatch(val)) return '';
    return _surroundQuotes.firstMatch(val)!.group(1)!;
  }

  String swallow(String line) => line.replaceAll(_keyword, '').trim();

  String unquote(String val) =>
      val.replaceFirstMapped(_surroundQuotes, (m) => m[2]!).trim();

  bool _has(Map<String, String> map, String key) =>
      map.containsKey(key) && map[key] != null;

  bool _isValid(String s) => s.isNotEmpty && s.contains('=');

  String? _tryPlatformEnv(String key) {
    if (!_has(Platform.environment, key)) return '';
    return Platform.environment[key];
  }
}
