import 'dart:typed_data';
import 'package:crypto/crypto.dart' as crypto;
import 'package:welltested/src/extensions/string_extensions.dart';

const NAMESPACE_OID = '6ba7b812-9dad-11d1-80b4-00c04fd430c8';

String uniqueCodeHex(String code, {String? namespace, bool compact = true}) {
  if (compact) code = code.shrink;

  return uuid(code, namespace: namespace);
}

String uuid(
  String name, {
  String? namespace,
}) {
  // Use provided namespace, or use whatever is decided by options.
  namespace = (namespace != null) ? namespace : NAMESPACE_OID;

  name = name;

  // Convert namespace UUID to Byte List
  var bytes = UuidParsing.parse(namespace);

  // Convert name to a list of bytes
  var nameBytes = <int>[];
  for (var singleChar in name.codeUnits) {
    nameBytes.add(singleChar);
  }

  // Generate SHA1 using namespace concatenated with name
  var hashBytes = crypto.sha1.convert([...bytes, ...nameBytes]).bytes;

  // per 4.4, set bits for version and clockSeq high and reserved
  hashBytes[6] = (hashBytes[6] & 0x0f) | 0x50;
  hashBytes[8] = (hashBytes[8] & 0x3f) | 0x80;

  return UuidParsing.unparse(hashBytes.sublist(0, 16));
}

class UuidParsing {
  /// Easy number -> hex conversion
  static final List<String> _byteToHex = List<String>.generate(256, (i) {
    return i.toRadixString(16).padLeft(2, '0');
  });

  /// Convert Hex String to Uint8List
  static Uint8List parseHexToBytes(String hex) {
    if (hex.length % 2 != 0) {
      throw ArgumentError('Invalid hex string');
    }
    if (hex.startsWith('0x')) {
      hex = hex.substring(2);
    }

    var bytes = Uint8List(hex.length ~/ 2);
    for (var i = 0; i < hex.length; i += 2) {
      bytes[i ~/ 2] = int.parse(hex.substring(i, i + 2), radix: 16);
    }
    return bytes;
  }

  static List<int> parse(
    String uuid, {
    List<int>? buffer,
    int offset = 0,
  }) {
    var i = offset, ii = 0;

    buffer = Uint8List(16);

    // Convert to lowercase and replace all hex with bytes then
    // string.replaceAll() does a lot of work that I don't need, and a manual
    // regex gives me more control.
    final regex = RegExp('[0-9a-f]{2}');
    for (Match match in regex.allMatches(uuid.toLowerCase())) {
      if (ii < 16) {
        var hex = uuid.toLowerCase().substring(match.start, match.end);
        buffer[i + ii++] = int.parse(hex, radix: 16);
      }
    }

    // Zero out any left over bytes if the string was too short.
    while (ii < 16) {
      buffer[i + ii++] = 0;
    }

    return buffer;
  }

  ///Parses the provided [uuid] into a list of byte values as a Uint8List.
  /// Can optionally be provided a [buffer] to write into and
  ///  a positional [offset] for where to start inputting into the buffer.
  /// Throws FormatException if the UUID is invalid. Optionally you can set
  /// [validate] to false to disable validation of the UUID before parsing.
  static Uint8List parseAsByteList(
    String uuid, {
    List<int>? buffer,
    int offset = 0,
  }) {
    return Uint8List.fromList(parse(
      uuid,
      buffer: buffer,
      offset: offset,
    ));
  }

  static String unparse(List<int> buffer, {int offset = 0}) {
    if (buffer.length - offset < 16) {
      throw RangeError('buffer too small: need 16: length=${buffer.length}'
          '${offset != 0 ? ', offset=$offset' : ''}');
    }
    var i = offset;
    return '${_byteToHex[buffer[i++]]}${_byteToHex[buffer[i++]]}'
        '${_byteToHex[buffer[i++]]}${_byteToHex[buffer[i++]]}-'
        '${_byteToHex[buffer[i++]]}${_byteToHex[buffer[i++]]}-'
        '${_byteToHex[buffer[i++]]}${_byteToHex[buffer[i++]]}-'
        '${_byteToHex[buffer[i++]]}${_byteToHex[buffer[i++]]}-'
        '${_byteToHex[buffer[i++]]}${_byteToHex[buffer[i++]]}'
        '${_byteToHex[buffer[i++]]}${_byteToHex[buffer[i++]]}'
        '${_byteToHex[buffer[i++]]}${_byteToHex[buffer[i++]]}';
  }
}
